;==============================================================
; WLA-DX banking setup
;==============================================================

.memorymap
defaultslot 0
slotsize $4000
slot 0 $0000
.endme

.rombankmap
bankstotal 1
banksize $4000
banks 1
.endro

;==============================================================
; SDSC tag and SMS rom header
;==============================================================

.sdsctag 5.0,"Video playback","Video playback demo","GliGli"

;==============================================================
; SMS defines
;==============================================================

.define VDPControl $bf
.define VDPData $be
.define VDPScanline $7e
.define VRAMWrite $4000
.define CRAMWrite $c000
.define MapperSlot0 $fffd
.define MapperSlot1 $fffe
.define MapperSlot2 $ffff
.define TileSize 32
.define TileMapSize 1536
.define TilePaletteSize 16

;==============================================================
; Program defines
;==============================================================

.define DblBufTileOffset 48

.macro SetVDPAddress
; Sets the VDP address
; Parameters: de = address
    ld a,e
    out (VDPControl),a
    ld a,d
    out (VDPControl),a
.endm

;==============================================================
; RAM variables
;==============================================================

.equ CurFrameIdx $c000
.equ CurVBlankIdx $c002

.equ LocalPalette $c0f0
.equ LocalTileMap $c100

.bank 0 slot 0
.org $0000
;==============================================================
; Boot section
;==============================================================
    di              ; disable interrupts
    im 1            ; Interrupt mode 1
    jp init         ; jump to mapper init

.org $0038
;==============================================================
; VDP int handler
;==============================================================
    ex af, af'

    ; VDP int ack
    in a, (VDPControl)

    ; CurVBlankIdx update
    ld a, (CurVBlankIdx)
    inc a
    ld (CurVBlankIdx), a

    ex af, af'
    ei
    reti

.org $0066
;==============================================================
; Pause button handler
;==============================================================
    retn

.org $0080
;==============================================================
; Mapper init
;==============================================================
init:
    ; This maps the first 48K of ROM to $0000-$BFFF
    ld de, $FFFC
    ld hl, init_tab
    ld bc, $0004
    ldir

    jp main

init_tab: ; Table must exist within first 1K of ROM
    .db $00, $00, $01, $02

;==============================================================
; Main program
;==============================================================
main:
    ld sp, $dff0

    ; Set up VDP registers

    ld hl,VDPInitData
    ld b,VDPInitDataEnd-VDPInitData
    ld c,VDPControl
    otir

    ; Clear VRAM

    ; 1. Set VRAM write address to $0000
    ld de, $0000 | VRAMWrite
    SetVDPAddress
    ; 2. Output 16KB of zeroes
    ld bc, $4000     ; Counter for 16KB of VRAM
-:  xor a
    out (VDPData), a ; Output to VRAM address, which is auto-incremented after each write
    dec bc
    ld a, b
    or c
    jr nz, -

    ; Dummy Sprite table
    ld de, $2000 | VRAMWrite
    SetVDPAddress
    ld a, $d0
    out (VDPData), a
    ld a, $20 << 1
    out (VDPControl), a
    ld a, $85
    out (VDPControl), a

    ; Init current frame idx
    xor a
    ld (CurFrameIdx), a
    ld (CurFrameIdx + $01), a
    ld (CurVBlankIdx), a
    ld (CurVBlankIdx + $01), a

    ; Turn screen on
    ld a, %1100000
;          ||||||`- Zoomed sprites -> 16x16 pixels
;          |||||`-- Doubled sprites -> 2 tiles per sprite, 8x16
;          ||||`--- Mega Drive mode 5 enable
;          |||`---- 30 row/240 line mode
;          ||`----- 28 row/224 line mode
;          |`------ VBlank interrupts
;          `------- Enable display
    out (VDPControl), a
    ld a, $81
    out (VDPControl), a

    ; Enable ints
    ei

frameLoad:
    ; Point on current frame index entry
    ld hl, (CurFrameIdx)
    add hl, hl
    ld de, VideoDataIndex + $02
    ex de, hl
    add hl, de
    add hl, de
    add hl, de

    ; Map slot 1 & 2 to current frame data
    ld a, (hl)
    inc hl
    inc hl
    ld (MapperSlot1), a
    inc a
    ld (MapperSlot2), a

    ; Load frame data address into hl and tiles offset into ix
    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl
    set 6, d; Add $4000
    ld a, (hl)
    ld ixl, a
    inc hl
    ld a, (hl)
    ld ixh, a
    ex de, hl

    ; Load palette if frame contains one
    ld de, LocalPalette
    ld a, (hl)
    inc hl
    cp $00
    jr z, NoFramePalette
    ld c, a
    ld b, $00
    ldir
NoFramePalette:

p1: ; Unpack tiles indexes and copy corresponding tiles to VRAM

        ; Set tiles VRAM start address
    ld a, (CurFrameIdx)
    and $01
    rra
    rra
    rra
    rra
    or ((DblBufTileOffset * TileSize) | VRAMWrite) >> 8
    ld d, a
    ld e, 0
    SetVDPAddress

        ; Load frame tile count
    ld e, (hl)
    inc hl

        ; Save current mapper slot
    ld a, (MapperSlot1)
    ld d, a

    push hl
TilesUploadLoop:
        ; Restore mapper slot
    ld a, d
    ld (MapperSlot1), a

        ; Get raw tile index (2 in 3 bytes)

    pop hl

    ld a, e
    and $01
    jr nz, SecondPart

FirstPart:
    ld c, (hl)
    inc hl
    ld a, (hl)
    jp SPEnd

SecondPart:
    ld a, (hl)
    inc hl
    ld c, (hl)
    inc hl
    rra
    rra
    rra
    rra

SPEnd:
    and $0f
    ld b, a
    push hl

        ; Add frame tile offset to tile index
    push ix
    pop hl
    add hl, bc

        ; Upper bits of tile index select a rom bank
    ld a, h
    sra a

    ld (MapperSlot1), a

        ; Lower bits select an offset in that bank
    add hl, hl
    add hl, hl
    add hl, hl
    add hl, hl
    add hl, hl
    set 6, h ; Add $4000
    res 7, h

        ; Upload tile to VRAM (fast unrolled during VBlank, slower during display)
    ld c, VDPData

    in a, (VDPScanline)
    cp 192
    jr c, TilesUploadSlow
    cp 253
    jr nc, TilesUploadSlow

TilesUploadFast:
    .repeat TileSize
        outi
    .endr
    dec e
    jr nz, TilesUploadLoop
    jp TilesUploadEnd

TilesUploadSlow:
    ld b, TileSize * 2
-:  outi
    djnz -
    dec e
    jr z, TilesUploadEnd
    jp TilesUploadLoop

TilesUploadEnd:
        ; Restore mapper slot
    ld a, d
    ld (MapperSlot1), a
    pop hl
    
    ; Unpack tilemap to RAM

    ld de, LocalTileMap
    ld b, 0

TilemapUnpack:
        ; load indicator
    ld a, (hl)
    inc hl

        ; skip or copy?
    cp 128
    jr c, TilemapCopy

TilemapSkip:
        ; skip part of tilemap we want to leave as-is
    sub 128
    jr z, TilemapUnpackEnd ; check for termination
    ld c, a
    ex de, hl
    add hl, bc
    ex de, hl
    jp TilemapUnpack

TilemapCopy:
        ; copy part of tilemap we want to update
    ld c, a
    ldir
    jp TilemapUnpack

TilemapUnpackEnd:

p2: ; Wait 4 VBlanks per frame (12.5 PAL fps)
-:  halt
    ld a, (CurVBlankIdx)
    and $03
    jr nz, -

p3: ; Upload local palette to VDP

    ld de, $0000 | CRAMWrite
    SetVDPAddress
    ld hl, LocalPalette
    ld c, VDPData
    .repeat TilePaletteSize
        outi
    .endr
    
    ; Upload local tilemap to VDP

    ld de, $0000 | VRAMWrite
    SetVDPAddress
    ld hl, LocalTileMap
    ld a, (CurFrameIdx)
    and $01
    .repeat TileMapSize / 2
        outi
        out (VDPData), a
    .endr

p4: ; Advance to next frame
    ld bc, (CurFrameIdx)
    inc bc
    ld hl, (VideoDataIndex)

    sbc hl, bc
    jr nz, +

    ; If we're past last frame, rewind to first frame
    ld bc, 0

+:
    ld (CurFrameIdx), bc
    jp frameLoad

;==============================================================
; Data
;==============================================================

; VDP initialisation data
VDPInitData:
.db $04,$80,$00,$81,$f1,$82,$ff,$85,$ff,$86,$ff,$87,$00,$88,$00,$89,$ff,$8a
VDPInitDataEnd:

VideoDataIndex:
.incbin "tiled/index.bin"
